from termcolor import cprint, colored

_DEBUG = False

_COLORS = ["light_red", "green", "yellow", "blue", "light_magenta"]


def set_debug(enabled: bool) -> None:
    global _DEBUG
    _DEBUG = enabled


def message(author: str, text: str, tags: list[str] | None = None) -> None:
    author = author.upper()
    idx = sum(ord(char) for char in author) % len(_COLORS)
    author_color = _COLORS[idx]
    title = author + (f" [{', '.join(tags)}]" if tags else "")
    cprint(colored(title, author_color, None, []))
    print(text)
    print("")


def info(text: str) -> None:
    cprint(text, "cyan")


def debug(text: str, indent="    ") -> None:
    if _DEBUG:
        indented = "\n".join([f"{indent}{x}" for x in text.splitlines()])
        cprint(indented, "grey")



from typing import Any, Dict
from ..http import http_request
from .settings import Settings


class ImagePrompt:
    def __init__(self, settings: Settings) -> None:
        """Initialize the prompt."""
        self.generations_url = 'https://api.openai.com/v1/chat/completions'
        self.settings = settings
        self.headers = {
            'authorization': f'Bearer {self.settings.OPENAI_API_KEY}',
            'openai-organization': self.settings.OPENAI_ORG_ID,
        }
        self.size = self.settings.OPENAI_IMAGE_SIZE

    def send_message(
        self, prompt: str, n: int = 1, size: str = ''
    ) -> Dict[str, Any]:
        """Send a message to the dialog and return the reply."""
        return http_request(
            url=self.generations_url, method='POST', headers=self.headers,
            json={'prompt': prompt, 'size': (size or self.size), 'n': n}
        )



import argparse
import time
from typing import List
from pathlib import Path
from ..http import http_request
from .settings import Settings
from .prompt import Prompt


def save_result(
    key: str, message: str, urls: List[str] = [],
    root_dir=Path('./history')
) -> None:
    """Save the generated result to the history directory."""
    chat_dir = root_dir / f'{key}-image'
    chat_dir.mkdir(parents=True, exist_ok=True)
    message_path = chat_dir / '0.txt'
    with open(message_path, "w") as file:
        file.write(message)
    for i, url in enumerate(urls):
        image_path = chat_dir / f'generated_{i}.png'
        print(url, '->', image_path)
        data = http_request('GET', url)
        with open(image_path, "wb") as file:
            file.write(data)


def main():
    """Generate images from a prompt message."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-n', '--images-number', metavar='IMAGES_NUMBER',
        type=int, default=1, required=False
    )
    parser.add_argument(
        '-s', '--image-size', metavar='IMAGE_SIZE',
        type=str, default='', required=False
    )
    args = parser.parse_args()
    #
    key = hex(int(time.time()))[2:]
    print(f"Prompt key: {key}")
    #
    try:
        message = input("Prompt for image by ChatGPT: ")
        message = message.strip()
    except (KeyboardInterrupt, EOFError):
        return
    #
    settings = Settings()
    prompt = Prompt(settings)
    result = prompt.send_message(
        message, n=args.images_number, size=args.image_size
    )
    #
    urls = [item['url'] for item in result.get('data', [])]
    save_result(key, message, urls, root_dir=Path(settings.HISTORY_ROOT))


if __name__ == '__main__':
    main()



from .base import BaseLexicon
from .prompts import (
    INSTRUCTIONAL_PROMPT,
    OUTPUT_FORMAT_PROMPT,
    KNOWN_ACTIONS_PROMT,
    CHECK_COMPLETION_PROMPT,
    INITIALIZER_REPLY_EXAMPLE,
)


class GeneralLexicon(BaseLexicon):
    role_model_prompt: str = ""

    def initializer_prompt(self) -> str:
        return (
            f"{self.role_model_prompt}\n\n"
            f"{INSTRUCTIONAL_PROMPT}\n\n"
            f"{OUTPUT_FORMAT_PROMPT}\n\n"
            f"{KNOWN_ACTIONS_PROMT}\n\n"
            f"{CHECK_COMPLETION_PROMPT}\n\n"
        ).strip()

    def initializer_history(self) -> list:
        prompt = self.initializer_prompt()
        reply = INITIALIZER_REPLY_EXAMPLE
        return [
            {'role': 'user', 'content': prompt},
            {'role': 'assistant', 'content': reply},
        ]



from golemgpt.actions import CODEX_ACTIONS
from golemgpt.settings import Verbosity
from golemgpt.utils import console
from golemgpt.utils.exceptions import AlignAcionsError
from golemgpt.cognitron.openai import OpenAIToolsCognitron
from .base import BaseCodex


class ReasonableCodex(BaseCodex):
    _cognitron: OpenAIToolsCognitron | None = None

    _name: str = "Reasonable-Codex"

    check_actions_depth = 3

    prescription_prompt = (
        "Is the following action plan harmful? "
        "NOTES: Consider asking user input as safe. "
        "Consider running known scripts as safe if aligned with goals. "
        "Be reasonable but not paranoid. "
    )

    message_ok = "Action looks OK."

    @property
    def cognitron(self) -> OpenAIToolsCognitron:
        if not self._cognitron:
            return OpenAIToolsCognitron(
                settings=self.settings,
                memory=self.memory,
                actions=CODEX_ACTIONS,
                name=self.name,
                verbosity=self.settings.VERBOSITY_CODEX,
            )
        return self._cognitron

    @property
    def name(self) -> str:
        return self._name

    def is_job_finished(self, actions: list[dict]) -> bool:
        if len(actions) == 1:
            return next(iter(actions[0].keys())) == "finish_job"
        return False

    def align_actions_prompt(self, actions: list[dict]) -> tuple[str, bool]:
        actions_summary = ""
        for item in actions:
            name = next(iter(item.keys()))
            actions_summary += f"\n{name}:\narguments = {item[name]}\n"
        prompt = f"{self.prescription_prompt}"
        prompt += f"ACTIONS: {actions_summary}"
        return self.cognitron.lexicon.yesno_prompt(prompt), False

    def align_actions(self, actions: list[dict]) -> bool:
        """Check if the actions are reasonably safe to perform."""

        if self.is_job_finished(actions):
            return True

        check_actions = actions[: self.check_actions_depth]
        prompt, answer_expected = self.align_actions_prompt(check_actions)

        if self.cognitron.verbosity >= Verbosity.VERBOSE:
            console.message(self.name, prompt, tags=["prompt"])

        answer_real = self.cognitron.ask_yesno(prompt)
        if answer_expected == answer_real:
            console.message(self.name, self.message_ok, tags=["align"])
            return True

        raise AlignAcionsError(prompt)



import json
from typing import Callable

from golemgpt.lexicon import BaseLexicon, Reply
from golemgpt.memory import BaseMemory
from golemgpt.settings import Settings


class BaseCognitron:
    DEFAULT_NAME = "Cognitron"

    LEXICON_CLASS = BaseLexicon

    def __init__(
        self,
        settings: Settings,
        memory: BaseMemory,
        actions: dict[str, Callable],
        name: str = "",
    ) -> None:
        self.settings = settings
        self.memory = memory
        self.name = name or self.DEFAULT_NAME
        self.lexicon = self.LEXICON_CLASS()
        self.actions = actions

    def communicate(self, message: str, **options) -> Reply:
        """Communicate with the Cognitron and return the reply."""
        raise NotImplementedError()

    def plan_actions(self, prompt: str, attempt: int = 0) -> list[dict]:
        raise NotImplementedError()

    def ask_yesno(self, message: str, **options) -> bool:
        """Ask the Cognitron a yes/no question and return the reply."""
        reply = self.communicate(message)
        return self.lexicon.guess_yesno(reply)

    def get_last_message(self) -> str:
        """Get the last message from the dialog with the Cognitron."""
        message = self.memory.messages[-1]
        if message.get("tool_calls"):
            return json.dumps(message.get("tool_calls"))
        return message["content"]



from json import dumps as json_dumps
from golemgpt.utils import workpath


def write_file_action(filename: str, content: str, **kwargs) -> str:
    """Write a file and return its info."""
    path = workpath(filename)

    if not isinstance(content, str):
        content = json_dumps(content)

    with path.open("w") as file:
        file.write(content)

    return f'File "{filename}" is saved.'



from enum import IntEnum
from pathlib import Path
from pydantic import BaseSettings


class Verbosity(IntEnum):
    """Output verbosity level."""
    SILENT = 0
    COMPACT = 1
    NORMAL = 2
    VERBOSE = 3
    DEBUG = 4


class Settings(BaseSettings):
    GOLEM_DEBUG: bool = False
    OPENAI_API_KEY: str = ""
    OPENAI_ORG_ID: str = ""
    OPENAI_MODEL: str = "gpt-4-turbo"
    WORKDIR: Path = Path("workdir")
    VERBOSITY_MAIN: Verbosity = Verbosity.NORMAL
    VERBOSITY_CODEX: Verbosity = Verbosity.SILENT

    class Config:
        env_file = ".env"



from golemgpt.actions import ALL_KNOWN_ACTIONS
from golemgpt.codex import BaseCodex
from golemgpt.codex import ReasonableCodex
from golemgpt.runners import JustDoRunner
from golemgpt.settings import Settings
from golemgpt.memory import BaseMemory
from golemgpt.utils import console, genkey
from golemgpt.cognitron.base import BaseCognitron

from golemgpt.cognitron.openai import OpenAIToolsCognitron
from golemgpt.utils.exceptions import (
    GolemError,
    AlignAcionsError,
    JobFinished,
    JobRejected,
    ParseActionsError,
    PathRejected,
)

DEFAULT_NAME = "Golem-GPT"

DEFAULT_RETRY_PLAN_ATTEMPTS = 3


class GeneralGolem:
    cognitron_class = OpenAIToolsCognitron
    codex_class = ReasonableCodex
    runner_class = JustDoRunner

    def __init__(
        self,
        *,
        name=DEFAULT_NAME,
        goals: list[str],
        job_key: str,
        memory: BaseMemory,
        settings: Settings,
    ) -> None:
        self.name = name
        self.action_plan = []
        self.job_key = job_key
        self.goals = goals
        self.memory = memory
        self.settings = settings
        self.runner = self.runner_class(settings)
        self.core = self.cognitron()

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.job_key})"

    def start_job(self) -> None:
        """Start the job."""
        goals_text = "\n".join(self.goals)
        console.info(f"Starting job: {self.job_key}")
        console.info(f"Goals:\n{goals_text}")

        self.initialize()

        outcome = self.core.lexicon.goal_prompt(self.goals[-1])
        while True:
            try:
                if outcome:
                    new_action_plan = self.plan_actions(outcome)
                    self.align_actions(new_action_plan)
                    self.action_plan = new_action_plan
                outcome = self.run_action()
            except JobFinished:
                break
            except JobRejected:
                break
            except PathRejected:
                break
            except AlignAcionsError as exc:
                # TODO: Implement a retry plan mechanism after misalignment
                console.info(f"Actions rejected: {exc}")
                break
            except GolemError as exc:
                outcome = str(exc)

        console.info(f"Job completed: {self.job_key}")

    def initialize(self) -> None:
        """Initialize the job state from memory or from scratch."""
        console.debug(f"Syncing job state with memory: {self.job_key}")
        self.memory.load(self.job_key)

        if self.memory.is_history_empty:
            self.memory.goals = self.goals
            iniital_history = self.core.lexicon.initializer_history()
            for message in iniital_history:
                self.memory.messages.append(message)
        else:
            self.goals = self.memory.goals
            assert self.goals

        self.memory.save()

    def cognitron(self, **options) -> BaseCognitron:
        """Return a new Cognitron instance."""
        assert self.settings, "Error: settings must be initialized first."
        assert self.memory, "Error: memory must be initialized first."
        key = f"{self.job_key}/{genkey()}"
        memory = self.memory.spawn(key)
        return self.cognitron_class(
            settings=self.settings,
            memory=memory,
            actions=ALL_KNOWN_ACTIONS,
            verbosity=self.settings.VERBOSITY_MAIN,
            **options,
        )

    def codex(self, **options) -> BaseCodex:
        """Return a new Codex instance."""
        assert self.settings, "Error: settings must be initialized first."
        assert self.memory, "Error: memory must be initialized first."
        key = f"{self.job_key}/{genkey()}"
        return self.codex_class(
            settings=self.settings,
            memory=self.memory.spawn(key)
        )  # fmt: skip

    def run_action(self) -> str:
        """Run the next action in the plan."""
        if not self.action_plan:
            raise JobFinished()
        action_item = self.action_plan.pop(0)
        action, result = self.runner(action_item, golem=self)
        if not result:
            return ""
        return self.core.lexicon.action_result_prompt(action, result)

    def plan_actions(self, prompt: str, attempt: int = 0) -> list[dict]:
        """Generate an action plan based on the prompt."""
        try:
            action_plan = self.core.plan_actions(prompt)
        except ParseActionsError:
            # TODO: Ingest format reminder message to the memory and retry,
            # check `lexicon.remind_format_prompt()`.
            raise
        return action_plan

    def align_actions(self, action_plan: list[dict]) -> list[str]:
        """Align the action plan with the codex."""
        return self.codex().align_actions(action_plan)



from json import load as json_load, dump as json_dump
from pathlib import Path
from .base import BaseMemory


class LocalFilesMemory(BaseMemory):
    def __init__(self, root_dir: Path) -> None:
        super().__init__()
        self.config['root_dir'] = root_dir

    @property
    def root_dir(self) -> Path:
        return self.config['root_dir']

    def get_path(self, filename: str) -> Path:
        assert self.key
        path = self.root_dir
        for key_part in self.key.split('/'):
            path = path / key_part
        return path / filename

    def spawn(self, key: str) -> 'LocalFilesMemory':
        """Spawn a fresh memory instance."""
        isinstance = LocalFilesMemory(self.root_dir)
        isinstance.load(key)
        return isinstance

    def load_file(self, filename: str, default: object) -> list:
        """Load JSON data from a local file."""
        path = self.get_path(filename)
        if path.exists():
            with path.open() as file:
                return json_load(file)
        return default

    def save_file(self, filename: str, data: object) -> None:
        """Save JSON data to a local file."""
        path = self.get_path(filename)
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open('w') as file:
            json_dump(data, file, indent=2, ensure_ascii=False)

    def load_messages(self) -> list:
        """Load the chat history from a local file."""
        return self.load_file('messages.json', [])

    def save_messages(self) -> None:
        """Save the chat history to a local file."""
        self.save_file('messages.json', self.messages)

    def load_goals(self) -> list:
        """Load the goals from a local file."""
        return self.load_file('job.json', {}).get('goals', [])

    def save_goals(self) -> None:
        """Save the goals to a local file."""
        return self.save_file('job.json', {'goals': self.goals})



from re import compile as re_compile, DOTALL

from pydantic import BaseModel

from golemgpt.utils.exceptions import UnknownReplyFormat


class Reply(BaseModel):
    """General representation of reploy from Cognitron."""

    text: str = ""
    actions: list = []


class BaseLexicon:
    # Not-a-JSON naive preamble regex:
    naive_preamble_re = re_compile(r'([^\[\{]*)', DOTALL)

    def initializer_prompt(self) -> str:
        return ""

    def goal_prompt(self, goal: str) -> str:
        tools_hint = (
            "NOTES: Use actions provided. "
            "Ask credentials from user if needed, be specific for which "
            "API endpoint you'll use it (exact address needed)."
        )
        goal = goal.strip().rstrip('.')
        return f"The goal is: {goal}.\n\n{tools_hint}"

    def yesno_prompt(self, question: str) -> str:
        return f"Answer 'yes' or 'no'. {question}"

    def action_result_prompt(self, action: str, result: str) -> str:
        return f"Completed '{action}' with result:\n{result}"

    def remind_format_prompt(self) -> str:
        return "If finished, just use a single 'finish_job' action."

    def guess_finish_prompt(self, reply: str) -> str:
        preamble = self.find_preamble(reply)
        statement = preamble or reply
        return (
            "Does the following mean current job is finished "
            "(optional ask to start a new one)?"
            f"\n\n{statement}"
        )

    def guess_yesno(self, reply: Reply) -> bool:
        """Guess if the reply is yes or no."""
        if not reply.text:
            raise UnknownReplyFormat(f"{reply}")
        return reply.text.lower().startswith("yes")

    def initializer_history(self) -> list:
        prompt = self.initializer_prompt()
        return [{"role": "user", "content": prompt}]

    def find_preamble(self, reply: str) -> str:
        preamble = self.naive_preamble_re.search(reply)
        if preamble:
            return preamble.group()
        return ""



import inspect
from json import loads as json_loads, JSONDecodeError
from typing import Any, Callable

from golemgpt.lexicon import BaseLexicon, GeneralLexicon, Reply
from golemgpt.memory import BaseMemory
from golemgpt.settings import Settings, Verbosity
from golemgpt.utils import console
from golemgpt.utils.exceptions import (
    ParseActionsError,
    UnknownAction,
    UnknownReplyFormat,
)
from golemgpt.utils.http import http_request
from .base import BaseCognitron


class OpenAITextCognitron(BaseCognitron):
    """
    Cognitron powered by OpenAI API with plain text communication.

    Response format and actions list are not enforced but provided with
    initial prompts. Extra heuristics required to process outputs, e.g.
    while converting response to the list of actions.

    Cons: results are less predictable, harder to manage, and debug.
    Pros: approach is more flexible and could be applied to other models
    that does not support explicit tools and output formats definition.

    """

    DEFAULT_NAME = "OpenAI-Text"
    MAX_TOKENS = 0
    TEMPERATURE = 0.1
    COMPLETIONS_URL = "https://api.openai.com/v1/chat/completions"
    LEXICON_CLASS = GeneralLexicon

    def __init__(
        self,
        settings: Settings,
        memory: BaseMemory,
        verbosity: Verbosity = Verbosity.NORMAL,
        **options
    ) -> None:
        super().__init__(settings, memory, **options)
        self.headers = {
            "authorization": f"Bearer {settings.OPENAI_API_KEY}",
            "openai-organization": settings.OPENAI_ORG_ID,
        }
        self.model = settings.OPENAI_MODEL
        self.memory = memory
        self.verbosity = verbosity

    def communicate(self, message: str, **options) -> Reply:
        """Communicate with the OpenAI and return the reply."""
        max_tokens = options.get("max_tokens", self.MAX_TOKENS)
        temperature = options.get("temperature", self.TEMPERATURE)
        messages = self.memory.messages.copy()
        messages.append({"role": "user", "content": message})
        payload = {
            "model": self.model,
            "messages": messages,
            "temperature": temperature,
        }
        if max_tokens:
            payload.update({"max_tokens": max_tokens})

        result = http_request(
            url=self.COMPLETIONS_URL,
            method="POST",
            headers=self.headers,
            json=payload,
        )
        reply = result["choices"][0]["message"]

        # Update history
        messages.append(reply)
        self.memory.messages = messages
        self.memory.save()

        # Print console output
        reply_text = self.get_last_message()

        if self.verbosity >= Verbosity.COMPACT:
            console.message(self.name, reply_text, tags=["reply"])

        return Reply(text=reply_text)

    def plan_actions(self, prompt: str, attempt: int = 0) -> list[dict]:
        """
        Plan actions based on the prompt using naive actions parsing.

        As we communicate the expected format in we'll try to parse actions
        from JSON response. It also could contain non-JSON preamble, which
        we try to detect and truncate.

        """
        if self.verbosity >= Verbosity.COMPACT:
            console.message(self.name, prompt, tags=["prompt"])

        reply = self.communicate(prompt)
        reply_text = reply.text
        console.debug(f"Parse plan:\n{reply_text}\n")

        # Remove not-a-JSON preamble, sometimes JSON is
        # prepended with some text, like "Here is your JSON:"
        preamble = self.lexicon.find_preamble(reply_text)
        if preamble:
            reply_text = reply_text[len(preamble) :]
            console.debug(f"Parse plan (trunc.):\n{reply_text}\n")

        reply_text = reply_text.strip()
        if not reply_text:
            raise ParseActionsError("JSON not found", reply_text)

        try:
            parsed = json_loads(reply_text)
        except JSONDecodeError as exc:
            raise ParseActionsError(exc.msg, reply_text) from exc

        if isinstance(parsed, dict):
            parsed = [parsed]

        return parsed


class OpenAIWithToolsLexicon(BaseLexicon):
    def initializer_prompt(self) -> str:
        return (
            "You are smart bot that can fulfill high-level tasks "
            "using API calls and other utilities."
        )

    def goal_prompt(self, goal: str) -> str:
        tools_hint = (
            "NOTES: Use tools provided. Prefer public APIs that do not "
            "require credentials. Ask credentials from user if needed, "
            "be specific for which API endpoint you'll use it (exact "
            "address needed)."
        )
        goal = goal.strip().rstrip(".")
        return f"The goal is: {goal}.\n\n{tools_hint}"

    def yesno_prompt(self, question: str) -> str:
        return f"Reply 'yes' or 'no'. {question}"

    def action_result_prompt(self, action: str, result: str) -> str:
        return f'Completed "{action}" with result:\n{result}'

    def guess_yesno(self, reply: Reply) -> bool:
        """Define by actions in the reply if it's yes or no ."""

        for item in reply.actions:
            name, args = self.parse_tool_call(item)
            if name == "reply_yes_or_no":
                return str(args.get("answer")).lower() in {"yes"}

        raise UnknownReplyFormat(f"{reply}")

    def parse_tool_call(self, item: dict) -> tuple[str, dict]:
        args = {}
        name = ""

        try:
            name = item["function"]["name"]
        except KeyError:
            if self.verbosity >= Verbosity.COMPACT:
                console.info(f"Unknown action: {item}")

        try:
            args = json_loads(item["function"]["arguments"])
        except KeyError:
            pass

        return (name, args)


class OpenAIToolsCognitron(BaseCognitron):
    """
    Cognitron powered by OpenAI API with 'tools' feature.

    Instead of trying to parse action plan naively from plain JSON we
    provide the list of tools during communication.

    The main difference from `OpenAINaiveCognitron` is that we always
    get only one next action instead of multi-step action plan.

    """

    DEFAULT_NAME = "OpenAI-Tools"
    MAX_TOKENS = 0
    TEMPERATURE = 0.1
    COMPLETIONS_URL = "https://api.openai.com/v1/chat/completions"
    LEXICON_CLASS = OpenAIWithToolsLexicon

    def __init__(
        self,
        settings: Settings,
        memory: BaseMemory,
        verbosity: Verbosity = Verbosity.NORMAL,
        **options
    ) -> None:
        super().__init__(settings, memory, **options)
        self.headers = {
            "authorization": f"Bearer {settings.OPENAI_API_KEY}",
            "openai-organization": settings.OPENAI_ORG_ID,
        }
        self.model = settings.OPENAI_MODEL
        self.tools = self.parse_actions_to_tools(self.actions)
        self.verbosity = verbosity

    @classmethod
    def parse_actions_to_tools(
        cls, actions: dict[str, Callable]
    ) -> list[dict[str, Any]]:
        """
        Parse known actions (callables with possible side-effects) to tools.

        'Tools' is the new feature of OpenAI's communication model. They can
        reply with the list of tools to use instead of plain text reply.
        """
        tools = []

        def _json_type(name: str) -> str:
            types_map = {
                "str": "string",
                "float": "number",
                "int": "number",
                "bool": "boolean",
                "dict": "object",
            }
            return types_map.get(name, "object")

        for key in actions:
            signature = inspect.signature(actions[key])
            if signature.parameters:
                parameters = {
                    "type": "object",
                    "properties": {
                        name: {"type": _json_type(arg.annotation.__name__)}
                        for name, arg in signature.parameters.items()
                    },
                }
            else:
                parameters = {}
            description = actions[key].__doc__ or " ".join(key.split("_"))
            tools.append(
                {
                    "type": "function",
                    "function": {
                        "name": key,
                        "description": description,
                        "parameters": parameters,
                    },
                }
            )
        return tools

    def plan_actions(self, prompt: str, attempt: int = 0) -> list[dict]:
        """
        Plan actions based on the prompt using with use of tools.

        We communicate the list of tools available via OpenAI call. We get
        the list of suggested calls in the reply. Actually it's a single-item
        list because that's how OpenAI's tools feature works currently.

        """
        if self.verbosity >= Verbosity.COMPACT:
            console.message(self.name, prompt, tags=["prompt"])
        reply = self.communicate(prompt)
        actions: list[dict] = []

        for item in reply.actions:
            name, args = self.lexicon.parse_tool_call(item)
            if name:
                actions.append({name: args})

        if not actions:
            raise UnknownAction(f"{reply}")

        return actions

    def communicate(self, message: str, **options) -> Reply:
        """Communicate with the OpenAI and return the reply."""
        max_tokens = options.get("max_tokens", self.MAX_TOKENS)
        temperature = options.get("temperature", self.TEMPERATURE)
        messages = self.memory.messages.copy()

        # If previous message was a tool call then we should provide
        # tool response in the specific format.
        tool_call_id = ""
        if messages:
            last_message = messages[-1]
            if "tool_calls" in last_message:
                last_tool_call = last_message["tool_calls"][0]
                tool_call_id = last_tool_call.get("id") or ""

        if tool_call_id:
            next_message = {
                "role": "tool",
                "content": message,
                "tool_call_id": tool_call_id,
            }
        else:
            next_message = {"role": "user", "content": message}
        messages.append(next_message)

        payload = {
            "model": self.model,
            "messages": messages,
            "temperature": temperature,
            "tools": self.tools,
            "tool_choice": "required",
        }
        if max_tokens:
            payload.update({"max_tokens": max_tokens})

        result = http_request(
            url=self.COMPLETIONS_URL,
            method="POST",
            headers=self.headers,
            json=payload,
        )
        reply = result["choices"][0]["message"]
        actions = reply["tool_calls"]

        # Update history
        messages.append(reply)
        self.memory.messages = messages
        self.memory.save()

        # Print console output
        reply_text = ""
        for item in actions:
            arguments = json_loads(item["function"]["arguments"])
            arguments_text = " | ".join(
                [f"{k}={v}" for k, v in arguments.items()]
            )
            reply_text += f'- {item["function"]["name"]}({arguments_text})'
        if self.verbosity >= Verbosity.COMPACT:
            console.message(self.name, reply_text.strip(), tags=["reply"])

        return Reply(actions=actions)



import os
import platform


def get_os_details_action(**kwargs) -> str:
    return {
        'os': os.name,
        'platform': platform.system(),
        'release': platform.release(),
        'version': platform.version(),
        'architecture': platform.machine(),
    }



# from .ask_google import ask_google_action
from .ask_human_input import ask_human_input_action
from .delegate_job import delegate_job_action
from .explain import explain_action
from .finish_job import finish_job_action
from .get_local_date import get_local_date_action
from .get_os_details import get_os_details_action
from .http_download import http_download_action
from .read_file import read_file_action
from .reject_job import reject_job_action
from .run_script import run_script_action
from .summarize_file import summarize_file_action
from .write_file import write_file_action

__all__ = [
    "ALL_KNOWN_ACTIONS",
]


def reply_yes_or_no(answer: str) -> bool:
    return answer.lower().startswith("yes")


ALL_KNOWN_ACTIONS = {
    "http_download": http_download_action,
    #
    "read_file": read_file_action,
    "summarize_file": summarize_file_action,
    "write_file": write_file_action,
    #
    "get_local_date": get_local_date_action,
    "get_os_details": get_os_details_action,
    #
    "run_script": run_script_action,
    #
    "explain": explain_action,
    #
    "ask_human_input": ask_human_input_action,
    "delegate_job": delegate_job_action,
    "reject_job": reject_job_action,
    "finish_job": finish_job_action,
    #
    "reply_yes_or_no": reply_yes_or_no,
}


CODEX_ACTIONS = {"reply_yes_or_no": reply_yes_or_no}



from subprocess import PIPE, Popen
from golemgpt.utils import workpath


def run_script_action(name: str, **kwargs) -> str:
    """Run a script and return its output."""
    path = workpath(name)

    if not path.exists():
        return (
            f"Script {name} does not exist, please write if first. "
            "Use a proper extension like .sh or .py for saved file."
        )

    if name.endswith('.py'):
        process = Popen(['python', name], stdout=PIPE, stderr=PIPE)
    elif name.endswith('.sh'):
        process = Popen(['sh', name], stdout=PIPE, stderr=PIPE)
    else:
        return f"Rejected: script {name} is not a Bash or Python script."

    output, error = process.communicate()
    is_ok = (process.returncode == 0)

    if isinstance(output, bytes):
        output = output.decode("utf-8")

    if isinstance(error, bytes):
        error = error.decode("utf-8")

    return (
        f"Script {name} executed successfully. Output:\n{output}"
        if is_ok else
        f"Script {name} failed. Error:\n{error}"
    )



from json import JSONDecodeError


class GolemError(Exception):
    """Base class for all Golem errors."""


class JobFinished(GolemError):
    """Exception to be raised when the job is finished."""


class JobRejected(GolemError):
    """Exception to be raised when the job is rejected."""


class PathRejected(GolemError, ValueError):
    """Raised when a requested path is rejected."""

    def __init__(self, path: str, *args):
        self.path = path
        super().__init__(*args)


class UnknownReplyFormat(GolemError, RuntimeError):
    """Raised when reply format is unknown."""


class UnknownAction(GolemError, RuntimeError):
    """Raised when unknown action is requested."""

    def __init__(self, name: str):
        self.name = name
        super().__init__(f"Unknown action: {name}.")


class AlignAcionsError(GolemError):
    """Exception to be raised when the action plan is rejected."""


class ParseActionsError(GolemError, JSONDecodeError):
    """Raised when actions plan cannot be parsed."""

    def __init__(self, msg: str, doc: str, pos: int = 0):
        super().__init__(msg, doc, pos)



from golemgpt.utils import workpath
from .summarize_file import summarize_file_action

MAX_RAW_SIZE = 8096


def read_file_action(filename: str, **kwargs) -> str:
    """Read a file and return its content."""
    path = workpath(filename, check_exists=True)
    file_size = path.stat().st_size
    if file_size > MAX_RAW_SIZE:
        hint = kwargs.get('summazie_hint') or "essential data is needed"
        return summarize_file_action(
            filename, hint=hint, **kwargs
        )
        # raise PathRejected(
        #     f"Error, file '{filename}' is too large ({file_size} bytes)."
        # )
    with path.open("r") as file:
        return file.read()



from golemgpt.memory import BaseMemory
from golemgpt.settings import Settings


class BaseCodex:
    def __init__(
        self, settings: Settings, memory: BaseMemory
    ) -> None:
        self.memory = memory
        self.settings = settings

    def align_actions(self, action_plan: list) -> str:
        raise NotImplementedError()



from typing import Any, Dict, Optional, Union
from json import dumps as json_dumps, loads as json_loads
import urllib3

http = urllib3.PoolManager()


def _do_request(
    method: str, url: str, headers: Optional[Dict[str, str]] = None,
    json: Optional[Any] = None, **kwargs: Any
) -> urllib3.HTTPResponse:
    """Send an HTTP request helper."""
    if json:
        headers = headers or {}
        headers['content-type'] = 'application/json'
        kwargs['body'] = json_dumps(json)
    return http.request(
        method=method, url=url, headers=headers, **kwargs
    )


def http_request_streamed(
    method: str, url: str, headers: Optional[Dict[str, str]] = None,
    json: Optional[Any] = None, **kwargs: Any
) -> urllib3.HTTPResponse:
    """Send an HTTP request without preloading, i.e. streamed."""
    return _do_request(
        method=method, url=url, headers=headers, json=json,
        preload_content=False, **kwargs
    )


def http_download(
    method: str, url: str, path: str,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[Any] = None, **kwargs: Any
) -> urllib3.HTTPResponse:
    """Download file via streamed HTTP request."""
    response = http_request_streamed(
        method=method, url=url, headers=headers, json=json, **kwargs
    )
    with open(path, 'wb') as file:
        for chunk in response.stream():
            file.write(chunk)
    return response


def http_request(
        method: str, url: str, headers: Optional[Dict[str, str]] = None,
        json: Optional[Any] = None, **kwargs: Any
    ) -> Union[Dict[str, Any], bytes]:
    """Send an HTTP request and return parsed data."""
    response = _do_request(
        method=method, url=url, headers=headers, json=json, **kwargs
    )
    if response.status == 200:
        if response.headers['content-type'] == 'application/json':
            return json_loads(response.data)
        return response.data
    raise RuntimeError(f'HTTP request failed: {response.data}')



from pydantic import BaseModel


class BaseMemory(BaseModel):
    key: str = ""
    goals: list = []
    messages: list[dict] = []
    config: dict = {}

    def spawn(self, key: str) -> "BaseMemory":
        """Spawn a fresh memory instance."""
        cls = self.__class__
        instance = cls(config=self.config)
        instance.load(key)
        return instance

    @property
    def is_history_empty(self) -> bool:
        return len(self.messages) == 0

    def load(self, key: str) -> None:
        """Load the job configuration and history."""
        self.key = key
        self.messages = self.load_messages()
        self.goals = self.load_goals()

    def save(self) -> None:
        """Save the job configuration and history."""
        self.save_messages()
        self.save_goals()

    def load_messages(self) -> list[dict]:
        raise NotImplementedError()

    def save_messages(self) -> None:
        raise NotImplementedError()

    def load_goals(self) -> list:
        raise NotImplementedError()

    def save_goals(self) -> None:
        raise NotImplementedError()

    def print_all_messages(self):
        """Print the chat history."""
        if self.is_empty:
            print("No messages.")
            return
        for item in self.messages:
            self.print_message(item)

    def print_last_message(self):
        """Print the last message in the chat history."""
        if not self.is_empty:
            self.print_message(self.messages[-1])

    def print_message(self, item: dict) -> None:
        """Print the last message in the chat history."""
        role = item['role'].capitalize()
        print(f"{role}:")
        print("-" * (1 + len(role)))
        print(item["content"].strip())
        print("\n")



from golemgpt.utils import console
from golemgpt.utils.exceptions import JobRejected


def reject_job_action(**kwargs) -> str:
    """Reject a job."""
    console.debug(f"Rejecting job: {kwargs}")
    raise JobRejected()



from json import dumps as json_dumps


INSTRUCTIONAL_PROMPT = """
Write a pseudo-code defining a list of actions to execute, so bots could
understand that. You provide a list of actions, bots will execute them
and you will get results.

Bypass already executed actions, use results provided.
""".strip()

INITIALIZER_REPLY_EXAMPLE = json_dumps([
    {'ask_human_input': {'query': 'What is the goal?'}},
]).strip()

# - create_python_script: name, description, in_files, out_files
# - create_shell_script: name, description, in_files, out_files
# - delegate_job: goal, role, in_files, out_files

KNOWN_ACTIONS_PROMT = """
You can use the following actions, higher in the list means
the higher priority:

- get_os_details:
- get_local_date:
- read_file: filename, summazie_hint
- write_file: filename, content
- http_download: url, method, headers, body, out_filename
- run_script: name
- ask_human_input: query
- explain: comment
- reject_job: message
- finish_job: message

Always find a corresponding actions from the list above.
""".strip()

KNOWN_ROLES_PROMT = ""

# KNOWN_ROLES_PROMT = """
# When using a the "delegate_job" action, assume the following roles:

# - "coder" can create function with defined goal, inputs and outputs
# - "admin" can install software on current machine
# """.strip()

# - "devops" can create configuration files for deployment
# - "webdesigner" can create HTML page and save it to the file
# - "artist" can create image and save it to the file
# - "reviewer" can review the code or action and provide feedback

OUTPUT_FORMAT_PROMPT = """
Always use machine readable JSON for every response. Don't add preambles.
Don't be verbose. Use the "explain" action to add thoughts, thanks, etc.

Format:

[
{"action_1": {"param_1": "value_1"}},
{"write_file": {"filename": "hello.txt", "content": "Hello there."}}
]
""".strip()

CHECK_COMPLETION_PROMPT = """
Check if job is completed after each action and respond with "finish_job"
on success.
""".strip()



from golemgpt.utils import workpath
from .write_file import write_file_action

MAX_RAW_SIZE = 32768


def summarize_file_action(
    filename: str, hint: str, out_filename: str, **kwargs
) -> str:
    # TODO: Better way to import for type checking
    from golemgpt.golems import GeneralGolem  # noqa
    golem = kwargs['golem']  # type: GeneralGolem
    # TODO: Split file into chunks for summarization
    path = workpath(filename, check_exists=True)
    with path.open("r") as file:
        content = file.read()
    content = content[:MAX_RAW_SIZE]
    cognitron = golem.cognitron()
    prompt = f"Summarize the text, use the hint '{hint}':\n\n{content}"
    reply = cognitron.communicate(prompt)
    return write_file_action(out_filename, reply)



from typing import Callable
from golemgpt.settings import Settings
from .justdo import JustDoRunner

__all__ = [
    'JustDoRunner',
]


def default_runner(settings: Settings) -> Callable:
    return JustDoRunner(settings)



from secrets import token_hex
from time import time
from typing import Optional, Union
from golemgpt.utils.misc import workpath
from golemgpt.utils.http import http_download

HTTP_REQUEST_PROMPT = """
Response saved to {out_filename} ({file_size} bytes).
""".strip()


def http_download_action(
    url: str, method: str = "GET", out_filename: str = "",
    headers: Optional[dict] = None,
    body: Union[str, list, dict] = None,
    **kwargs
) -> str:
    """Make an HTTP request and return its content."""

    if not out_filename:
        out_filename = f"out_{int(time())}_{token_hex(12)}.txt"

    try:
        path = workpath(out_filename)
    except ValueError:
        return f"File {out_filename} is outside of working dir."

    response = http_download(
        method=method, url=url, path=path, headers=headers,
        json=body if isinstance(body, (dict, list)) else None,
        body=body if isinstance(body, str) else None,
    )

    if response.status in (401, 403):
        return "HTTP request failed: maybe ask user for credentials?"

    file_size = path.stat().st_size
    return HTTP_REQUEST_PROMPT.format(
        out_filename=out_filename, file_size=file_size
    )
